;8-Bit to 7-Segment-Display converter
;display modes:
;  0) unsigned
;  1) signed
;  2) hexadecimal
;  3) inverted, unsigned
;  4) inverted, signed
;  5) inverted, hexadecimal
;ATTiny2313 @ 1 MHz internal R/C (lfuse=0x64, hfuse=0x9B)
;
;PORTS:
;  PB0-7 = input 0~7 (with internal pullups)
;  PD0-6 = segment a~g (active high)
;  PA0   = display 2 (middle), active high
;  PA1   = display 3 (right), active high
;  RESET = toggle display mode (push-button to ground)
;connect display 1 (left) cathode to ground through a diode (1N4001)

.include "../include/tn2313def.inc"

.def	temp = R16
.def	temp2 = R17
.def	inp = R18
.def	disp1 = R19
.def	disp2 = R20
.def	disp3 = R21
.def	zero_reg = R22
.def	mode = R23
.def	timeoutL = R24
.def	timeoutH = R25

.equ	SEG_A = 1
.equ	SEG_B = 2
.equ	SEG_C = 4
.equ	SEG_D = 8
.equ	SEG_E = 16
.equ	SEG_F = 32
.equ	SEG_G = 64

.equ	MINUS  = SEG_G
.equ	ZERO   = SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F
.equ	ONE    = SEG_B | SEG_C
.equ	TWO    = SEG_A | SEG_B | SEG_D | SEG_E | SEG_G
.equ	THREE  = SEG_A | SEG_B | SEG_C | SEG_D | SEG_G
.equ	FOUR   = SEG_B | SEG_C | SEG_F | SEG_G
.equ	FIVE   = SEG_A | SEG_C | SEG_D | SEG_F | SEG_G
.equ	SIX    = SEG_A | SEG_C | SEG_D | SEG_E | SEG_F | SEG_G
.equ	SEVEN  = SEG_A | SEG_B | SEG_C
.equ	EIGHT  = SEG_A | SEG_B | SEG_C | SEG_D | SEG_E | SEG_F | SEG_G
.equ	NINE   = SEG_A | SEG_B | SEG_C | SEG_D | SEG_F | SEG_G
.equ	HEX_A  = SEG_A | SEG_B | SEG_C | SEG_E | SEG_F | SEG_G
.equ	HEX_B  = SEG_C | SEG_D | SEG_E | SEG_F | SEG_G
.equ	HEX_C  = SEG_A | SEG_D | SEG_E | SEG_F
.equ	HEX_D  = SEG_B | SEG_C | SEG_D | SEG_E | SEG_G
.equ	HEX_E  = SEG_A | SEG_D | SEG_E | SEG_F | SEG_G
.equ	HEX_F  = SEG_A | SEG_E | SEG_F | SEG_G
.equ	HEX_h  = SEG_C | SEG_E | SEG_F | SEG_G
.equ	MODE_U = SEG_B | SEG_C | SEG_D | SEG_E | SEG_F
.equ	MODE_S = SEG_A | SEG_C | SEG_D | SEG_F | SEG_G
.equ	MODE_H = SEG_B | SEG_C | SEG_E | SEG_F | SEG_G
.equ	MODE_I = SEG_E | SEG_F

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
	
charset:
	.db ZERO, ONE, TWO, THREE, FOUR, FIVE, SIX, SEVEN
	.db EIGHT, NINE, HEX_A, HEX_B, HEX_C, HEX_D, HEX_E, HEX_F

;===============================================================================

reset:
	;stackpointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;ports
	ldi	temp, 0x00
	out	DDRB, temp
	ldi	temp, 0xFF
	out	PORTB, temp
	ldi	temp, 0x7F
	out	DDRD, temp
	ldi	temp, 0x00
	out	PORTD, temp
	ldi	temp, 0x03
	out	DDRA, temp
	ldi	temp, 0x04
	out	PORTA, temp
	
	;registers
	ldi	zero_reg, 0
	
	;load display mode from EEPROM
	ldi	temp, 42
	out	EEAR, temp
	sbi	EECR, EERE
	in	mode, EEDR
	cpi	mode, 6
	brlo	PC+3
	ldi	mode, 0
	rcall	mode_write
	
	;increment mode on external reset (long start-up delay = debouncing!)
	in	temp, MCUSR
	sbrs	temp, EXTRF
	rjmp	reset_end
	out	MCUSR, zero_reg
	inc	mode
	cpi	mode, 6
	brlo	PC+2
	ldi	mode, 0
	rcall	mode_write
reset_end:

	;display current mode setting
	ldi	disp3, 0 ;blank
	cpi	mode, 3
	brlo	PC+2
	ldi	disp3, MODE_I ;I = Inverted
	
	ldi	disp1, MODE_U ;U = Unsigned
	cpi	mode, 1
	brne	PC+2
	ldi	disp1, MODE_S ;S = Signed
	cpi	mode, 4
	brne	PC+2
	ldi	disp1, MODE_S ;S = Signed
	cpi	mode, 2
	brne	PC+2
	ldi	disp1, MODE_H ;H = Hexadecimal
	cpi	mode, 5
	brne	PC+2
	ldi	disp1, MODE_H ;H = Hexadecimal
	
	ldi	disp2, 0 ;blank
	ldi	timeoutL, LOW(333) ;1s timeout
	ldi	timeoutH, HIGH(333)
	
;-------------------------------------------------------------------------------
	
loop:
	cpi	timeoutL, 0
	cpc	timeoutH, timeoutL
	breq	calc
	sbiw	timeoutH:timeoutL, 1
	brne	calc_end
calc:
	ldi	disp1, 0
	ldi	disp2, 0
	in	temp, PINB
	cpi	mode, 3
	brlo	PC+2
	com	temp ;invert
	cpi	mode, 1
	breq	signed
	cpi	mode, 4
	breq	signed
	cpi	mode, 2
	breq	hex
	cpi	mode, 5
	breq	hex
	
unsigned:
	;unsigned
	cpi	temp, 100
	brlo	PC+3
	ldi	disp1, ONE
	subi	temp, 100
	cpi	temp, 100
	brlo	PC+3
	ldi	disp1, TWO
	subi	temp, 100
	rjmp	signed_end
	
signed:
	;signed
	cpi	temp, 128 ;positive if < 127
	brlo	unsigned
	neg	temp
	ldi	disp1, MINUS
	cpi	temp, 100
	brlo	signed_end
	subi	temp, 100
	ori	disp1, ONE
signed_end:
	
	;10's
div10:
	cpi	temp, 10
	brlo	div10_end
	subi	temp, 10
	inc	disp2
	rjmp	div10
div10_end:
	;1's
	mov	disp3, temp
	rjmp	encode
	
hex:
	;hexadecimal
	ldi	disp1, HEX_h ;display 'h' in front of hex number
	mov	disp2, temp
	swap	disp2
	andi	disp2, 0x0F
	mov	disp3, temp
	andi	disp3, 0x0F

	;encode 10's and 1's to 7-segment
encode:
	ldi	ZL, LOW(charset*2)
	ldi	ZH, HIGH(charset*2)
	add	ZL, disp2
	adc	ZH, zero_reg
	lpm	disp2, Z
	ldi	ZL, LOW(charset*2)
	ldi	ZH, HIGH(charset*2)
	add	ZL, disp3
	adc	ZH, zero_reg
	lpm	disp3, Z
	;zero blanking
	cpi	disp1, 0
	breq	PC+3
	cpi	disp1, MINUS
	brne	calc_end
	cpi	disp2, ZERO
	brne	calc_end
	mov	disp2, disp1 ;move minus (if present)
	ldi	disp1, 0
calc_end:
	
	;display 1
	out	PORTD, disp1
	rcall	delay
	out	PORTD, zero_reg
	rjmp	PC+1
	
	;display 2
	sbi	PORTA, 0
	rjmp	PC+1
	out	PORTD, disp2
	rcall	delay
	out	PORTD, zero_reg
	rjmp	PC+1
	cbi	PORTA, 0
	
	;display 3
	sbi	PORTA, 1
	rjmp	PC+1
	out	PORTD, disp3
	rcall	delay
	out	PORTD, zero_reg
	rjmp	PC+1
	cbi	PORTA, 1
	
	rjmp	loop
	
;===============================================================================

delay:
	;delay ~1ms
	ldi	temp, 0
delay_loop:
	nop
	dec	temp
	brne	delay_loop
	ret
	
;--------------------

mode_write:
	;write current mode to eeprom
	ldi	temp, 42
	out	EEAR, temp
	out	EEDR, mode
	sbi	EECR, EEMPE
	sbi	EECR, EEPE
	ret
	